import copy
import json
import logging
from http.client import HTTPResponse
from typing import Union
from urllib.error import HTTPError
from urllib.parse import urlencode, quote
from urllib.request import Request, urlopen

from .errors import SCIMApiError, SCIMError
from .group import Group
from .groups import Groups
from .request import SCIMRequest
from .response import SCIMResponse
from .service_provider_configs import ServiceProviderConfigs
from .user import User
from .users import Users


class SCIMClient:
    production_base_url = "https://api.slack.com/scim/v1"
    _logger = logging.getLogger(__name__)
    schema_values = ["urn:scim:schemas:core:1.0", "urn:scim:schemas:extension:enterprise:1.0"]

    def __init__(
        self,
        token: str,
        base_url: str = production_base_url
    ):
        self.token: str = token
        self.base_url: str = base_url

    # ----------------------------------------------
    # User Management
    # ----------------------------------------------

    def create_user(
        self,
        user: Union[dict, User]
    ) -> User:
        req = SCIMRequest(
            token=self.token,
            http_method="POST",
            url=f"{self.base_url}/Users",
            json_body=self._to_user_dict(user),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return User.from_dict(json.loads(resp.body))
        else:
            raise SCIMApiError.from_response(resp)

    def patch_user(
        self,
        id: str,
        user: Union[dict, User]
    ) -> User:
        id = self._ensure_user_id(id, user)
        req = SCIMRequest(
            token=self.token,
            http_method="PATCH",
            url=f"{self.base_url}/Users/{quote(id)}",
            json_body=self._to_user_dict(user),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return User.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def update_user(
        self,
        id: str,
        user: Union[dict, User],
    ) -> User:
        id = self._ensure_user_id(id, user)
        req = SCIMRequest(
            token=self.token,
            http_method="PUT",
            url=f"{self.base_url}/Users/{quote(id)}",
            json_body=self._to_user_dict(user),
        )
        resp = self.api_call(req)
        if resp.status == 200:
            return User.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def delete_user(
        self,
        id: str,
    ) -> User:
        req = SCIMRequest(
            token=self.token,
            http_method="DELETE",
            url=f"{self.base_url}/Users/{quote(id)}",
        )
        resp = self.api_call(req)
        if not resp.is_success():
            raise SCIMApiError.from_response(resp)

    def read_user(
        self,
        id: str,
    ) -> User:
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Users/{quote(id)}",
        )
        resp = self.api_call(req)
        if resp.is_success():
            return User.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def search_users(
        self,
        *,
        filter: str = None,
        count: int = None,
        start_index: int = None
    ) -> Users:
        query = {}
        if filter:
            query["filter"] = filter
        if count:
            query["count"] = count
        if start_index:
            query["startIndex"] = start_index
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Users",
            query_params=query
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Users.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def _ensure_user_id(self, id: str, user: User) -> str:
        if id:
            return id
        else:
            if user:
                return user.id
            else:
                raise SCIMError("User ID is missing for update_user call")

    def _to_user_dict(self, user: Union[dict, User]) -> dict:
        if user:
            user_dict: dict = user if isinstance(user, dict) else user.to_dict()
            user_dict["schemas"] = self.schema_values
            return user_dict
        else:
            return None

    # ----------------------------------------------
    # Group Management
    # ----------------------------------------------

    def create_group(
        self,
        group: Union[dict, Group]
    ) -> Group:
        req = SCIMRequest(
            token=self.token,
            http_method="POST",
            url=f"{self.base_url}/Groups",
            json_body=self._to_group_dict(group),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def patch_group(
        self,
        id: str,
        group: Union[dict, Group]
    ) -> Group:
        id = self._ensure_group_id(id, group)
        req = SCIMRequest(
            token=self.token,
            http_method="PATCH",
            url=f"{self.base_url}/Groups/{quote(id)}",
            json_body=self._to_group_dict(group),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def update_group(
        self,
        id: str,
        group: Union[dict, Group],
    ) -> Group:
        id = self._ensure_group_id(id, group)
        req = SCIMRequest(
            token=self.token,
            http_method="PUT",
            url=f"{self.base_url}/Groups/{quote(id)}",
            json_body=self._to_group_dict(group),
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def delete_group(
        self,
        id: str,
    ) -> Group:
        req = SCIMRequest(
            token=self.token,
            http_method="DELETE",
            url=f"{self.base_url}/Groups/{quote(id)}",
        )
        resp = self.api_call(req)
        if not resp.is_success():
            raise SCIMApiError.from_response(resp)

    def read_group(
        self,
        id: str,
    ) -> Group:
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Groups/{quote(id)}",
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Group.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def search_groups(
        self,
        *,
        filter: str = None,
        count: int = None,
        start_index: int = None
    ) -> Groups:
        query = {}
        if filter:
            query["filter"] = filter
        if count:
            query["count"] = count
        if start_index:
            query["startIndex"] = start_index
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/Groups",
            query_params=query
        )
        resp = self.api_call(req)
        if resp.is_success():
            return Groups.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    def _ensure_group_id(self, id: str, group: Group) -> str:
        if id:
            return id
        else:
            if group:
                return group.id
            else:
                raise SCIMError("Group ID is missing for update_group call")

    def _to_group_dict(self, group: Union[dict, Group]) -> dict:
        if group:
            group_dict: dict = group if isinstance(group, dict) else group.to_dict()
            group_dict["schemas"] = self.schema_values
            return group_dict
        else:
            return None

    # ----------------------------------------------
    # ServiceProviderConfigs
    # ----------------------------------------------

    def get_service_provider_configs(self) -> ServiceProviderConfigs:
        # ServiceProviderConfigs
        req = SCIMRequest(
            token=self.token,
            http_method="GET",
            url=f"{self.base_url}/ServiceProviderConfigs",
        )
        resp = self.api_call(req)
        if resp.is_success():
            return ServiceProviderConfigs.from_dict(json.loads(resp.body)) if resp.body else None
        else:
            raise SCIMApiError.from_response(resp)

    # ----------------------------------------------
    # HTTP Client
    # ----------------------------------------------

    def api_call(self, api_request: SCIMRequest) -> SCIMResponse:
        http_method = api_request.http_method.upper()
        url = api_request.url
        if api_request.query_params:
            params = copy.copy(api_request.query_params)
            if http_method == "GET" and api_request.body_params:
                params.update(api_request.body_params)
            q = urlencode(params)
            url = url + (f"&{q}" if "?" in url else f"?{q}")
            pass

        headers = copy.copy(api_request.headers)
        headers["Authorization"] = f"Bearer {api_request.token}"
        if api_request.json_body:
            body: dict = self._to_non_null_dict(copy.copy(api_request.json_body))
            req_body: str = None if http_method == "GET" \
                else json.dumps(body).encode("utf-8")
            headers["Content-Type"] = "application/json;charset=utf-8"
        else:
            req_body: str = None if http_method == "GET" \
                else urlencode(api_request.body_params).encode("utf-8")
            headers["Content-Type"] = "application/x-www-form-urlencoded;charset=utf-8"

        try:
            http_request: Request = Request(
                method=http_method,
                url=url,
                data=req_body,
                headers=headers
            )
            self._debug_log_request(http_request, req_body)

            http_response: HTTPResponse = urlopen(http_request)
            charset: str = http_response.headers.get_content_charset()
            raw_body: bytes = http_response.read()
            resp_body: str = raw_body.decode(charset) if raw_body else None

            api_response = SCIMResponse(
                status=http_response.status,
                reason=http_response.reason,
                headers=http_response.headers,
                body=resp_body,
            )
            self._debug_log_completion(http_request, api_response)
            return api_response

        except HTTPError as e:
            # TODO: rate limit
            api_response = SCIMResponse(
                status=e.code,
                reason=e.reason,
                headers=e.headers,
                body=e.read().decode("utf-8"),
            )
            self._debug_log_completion(http_request, api_response)
            return api_response
        except Exception as e:
            self._logger.error(f"Failed to send a request to Slack SCIM API server: {e}")
            raise e

    def _to_non_null_dict(self, d: Union[dict, any]) -> dict:
        result = {}
        if isinstance(d, dict):
            for key, value in d.items():
                if value is None:
                    continue
                if isinstance(value, dict):
                    value = self._to_non_null_dict(value)
                    result[key] = value
                elif isinstance(value, list):
                    new_list = []
                    for v in value:
                        v = self._to_non_null_dict(v)
                        if v:
                            new_list.append(v)
                    result[key] = new_list
                else:
                    result[key] = value
            return result
        elif getattr(d, "to_dict", None):
            return d.to_dict()
        else:
            return d

    def _debug_log_request(self, req: Request, body: str):
        if self._logger.level <= logging.DEBUG:
            headers_part = "\n".join([
                f"{k}: (redacted)" if k.lower() == "authorization" else f"{k}: {v}"
                for k, v in req.headers.items()
            ])
            message = f"*** SCIM API Request ***\n" \
                      f"{req.method} {req.full_url}\n" \
                      f"{headers_part}\n\n" \
                      f"{body or ''}\n"
            self._logger.debug(message)

    def _debug_log_completion(self, req: Request, resp: SCIMResponse):
        if self._logger.level <= logging.DEBUG:
            headers_part = "\n".join([f"{k}: {v}" for k, v in req.headers.items()])
            message = f"*** SCIM API Response ***\n" \
                      f"{req.method} {req.full_url}\n" \
                      f"{resp.status} {resp.reason}\n" \
                      f"{headers_part}\n\n" \
                      f"{resp.body or ''}\n"
            self._logger.debug(message)



# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = users_from_dict(json.loads(json_string))

from typing import Optional, Any, List, TypeVar, Type, cast, Callable


T = TypeVar("T")


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


class UsersEmail:
    primary: Optional[bool]
    value: Optional[str]

    def __init__(self, primary: Optional[bool], value: Optional[str]) -> None:
        self.primary = primary
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'UsersEmail':
        assert isinstance(obj, dict)
        primary = from_union([from_bool, from_none], obj.get("primary"))
        value = from_union([from_str, from_none], obj.get("value"))
        return UsersEmail(primary, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["primary"] = from_union([from_bool, from_none], self.primary)
        result["value"] = from_union([from_str, from_none], self.value)
        return result


class Meta:
    created: Optional[str]
    location: Optional[str]

    def __init__(self, created: Optional[str], location: Optional[str]) -> None:
        self.created = created
        self.location = location

    @staticmethod
    def from_dict(obj: Any) -> 'Meta':
        assert isinstance(obj, dict)
        created = from_union([from_str, from_none], obj.get("created"))
        location = from_union([from_str, from_none], obj.get("location"))
        return Meta(created, location)

    def to_dict(self) -> dict:
        result: dict = {}
        result["created"] = from_union([from_str, from_none], self.created)
        result["location"] = from_union([from_str, from_none], self.location)
        return result


class Name:
    family_name: Optional[str]
    given_name: Optional[str]

    def __init__(self, family_name: Optional[str], given_name: Optional[str]) -> None:
        self.family_name = family_name
        self.given_name = given_name

    @staticmethod
    def from_dict(obj: Any) -> 'Name':
        assert isinstance(obj, dict)
        family_name = from_union([from_str, from_none], obj.get("familyName"))
        given_name = from_union([from_str, from_none], obj.get("givenName"))
        return Name(family_name, given_name)

    def to_dict(self) -> dict:
        result: dict = {}
        result["familyName"] = from_union([from_str, from_none], self.family_name)
        result["givenName"] = from_union([from_str, from_none], self.given_name)
        return result


class Photo:
    type: Optional[str]
    value: Optional[str]

    def __init__(self, type: Optional[str], value: Optional[str]) -> None:
        self.type = type
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'Photo':
        assert isinstance(obj, dict)
        type = from_union([from_str, from_none], obj.get("type"))
        value = from_union([from_str, from_none], obj.get("value"))
        return Photo(type, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["type"] = from_union([from_str, from_none], self.type)
        result["value"] = from_union([from_str, from_none], self.value)
        return result


class Address:
    country: Optional[str]
    locality: Optional[str]
    postal_code: Optional[str]
    primary: Optional[bool]
    region: Optional[str]
    street_address: Optional[str]

    def __init__(self, country: Optional[str], locality: Optional[str], postal_code: Optional[str], primary: Optional[bool], region: Optional[str], street_address: Optional[str]) -> None:
        self.country = country
        self.locality = locality
        self.postal_code = postal_code
        self.primary = primary
        self.region = region
        self.street_address = street_address

    @staticmethod
    def from_dict(obj: Any) -> 'Address':
        assert isinstance(obj, dict)
        country = from_union([from_str, from_none], obj.get("country"))
        locality = from_union([from_str, from_none], obj.get("locality"))
        postal_code = from_union([from_str, from_none], obj.get("postalCode"))
        primary = from_union([from_bool, from_none], obj.get("primary"))
        region = from_union([from_str, from_none], obj.get("region"))
        street_address = from_union([from_str, from_none], obj.get("streetAddress"))
        return Address(country, locality, postal_code, primary, region, street_address)

    def to_dict(self) -> dict:
        result: dict = {}
        result["country"] = from_union([from_str, from_none], self.country)
        result["locality"] = from_union([from_str, from_none], self.locality)
        result["postalCode"] = from_union([from_str, from_none], self.postal_code)
        result["primary"] = from_union([from_bool, from_none], self.primary)
        result["region"] = from_union([from_str, from_none], self.region)
        result["streetAddress"] = from_union([from_str, from_none], self.street_address)
        return result


class PhoneNumberElement:
    primary: Optional[bool]
    type: Optional[str]
    value: Optional[str]

    def __init__(self, primary: Optional[bool], type: Optional[str], value: Optional[str]) -> None:
        self.primary = primary
        self.type = type
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'PhoneNumberElement':
        assert isinstance(obj, dict)
        primary = from_union([from_bool, from_none], obj.get("primary"))
        type = from_union([from_str, from_none], obj.get("type"))
        value = from_union([from_str, from_none], obj.get("value"))
        return PhoneNumberElement(primary, type, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["primary"] = from_union([from_bool, from_none], self.primary)
        result["type"] = from_union([from_str, from_none], self.type)
        result["value"] = from_union([from_str, from_none], self.value)
        return result


class Group:
    display: Optional[str]
    value: Optional[str]

    def __init__(self, display: Optional[str], value: Optional[str]) -> None:
        self.display = display
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'Group':
        assert isinstance(obj, dict)
        display = from_union([from_str, from_none], obj.get("display"))
        value = from_union([from_str, from_none], obj.get("value"))
        return Group(display, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["display"] = from_union([from_str, from_none], self.display)
        result["value"] = from_union([from_str, from_none], self.value)
        return result


class Manager:
    pass

    def __init__(self, ) -> None:
        pass

    @staticmethod
    def from_dict(obj: Any) -> 'Manager':
        assert isinstance(obj, dict)
        return Manager()

    def to_dict(self) -> dict:
        result: dict = {}
        return result


class UrnScimSchemasExtensionEnterprise10:
    manager: Optional[Manager]

    def __init__(self, manager: Optional[Manager]) -> None:
        self.manager = manager

    @staticmethod
    def from_dict(obj: Any) -> 'UrnScimSchemasExtensionEnterprise10':
        assert isinstance(obj, dict)
        manager = from_union([Manager.from_dict, from_none], obj.get("manager"))
        return UrnScimSchemasExtensionEnterprise10(manager)

    def to_dict(self) -> dict:
        result: dict = {}
        result["manager"] = from_union([lambda x: to_class(Manager, x), from_none], self.manager)
        return result


class Resource:
    active: Optional[bool]
    addresses: Optional[List[Address]]
    display_name: Optional[str]
    emails: Optional[List[PhoneNumberElement]]
    external_id: Optional[str]
    groups: Optional[List[Group]]
    id: Optional[str]
    meta: Optional[Meta]
    name: Optional[Name]
    nick_name: Optional[str]
    phone_numbers: Optional[List[PhoneNumberElement]]
    photos: Optional[List[Photo]]
    profile_url: Optional[str]
    roles: Optional[List[PhoneNumberElement]]
    schemas: Optional[List[str]]
    timezone: Optional[str]
    title: Optional[str]
    urn_scim_schemas_extension_enterprise_10: Optional[UrnScimSchemasExtensionEnterprise10]
    user_name: Optional[str]

    def __init__(self, active: Optional[bool], addresses: Optional[List[Address]], display_name: Optional[str], emails: Optional[List[PhoneNumberElement]], external_id: Optional[str], groups: Optional[List[Group]], id: Optional[str], meta: Optional[Meta], name: Optional[Name], nick_name: Optional[str], phone_numbers: Optional[List[PhoneNumberElement]], photos: Optional[List[Photo]], profile_url: Optional[str], roles: Optional[List[PhoneNumberElement]], schemas: Optional[List[str]], timezone: Optional[str], title: Optional[str], urn_scim_schemas_extension_enterprise_10: Optional[UrnScimSchemasExtensionEnterprise10], user_name: Optional[str]) -> None:
        self.active = active
        self.addresses = addresses
        self.display_name = display_name
        self.emails = emails
        self.external_id = external_id
        self.groups = groups
        self.id = id
        self.meta = meta
        self.name = name
        self.nick_name = nick_name
        self.phone_numbers = phone_numbers
        self.photos = photos
        self.profile_url = profile_url
        self.roles = roles
        self.schemas = schemas
        self.timezone = timezone
        self.title = title
        self.urn_scim_schemas_extension_enterprise_10 = urn_scim_schemas_extension_enterprise_10
        self.user_name = user_name

    @staticmethod
    def from_dict(obj: Any) -> 'Resource':
        assert isinstance(obj, dict)
        active = from_union([from_bool, from_none], obj.get("active"))
        addresses = from_union([lambda x: from_list(Address.from_dict, x), from_none], obj.get("addresses"))
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        emails = from_union([lambda x: from_list(PhoneNumberElement.from_dict, x), from_none], obj.get("emails"))
        external_id = from_union([from_str, from_none], obj.get("externalId"))
        groups = from_union([lambda x: from_list(Group.from_dict, x), from_none], obj.get("groups"))
        id = from_union([from_str, from_none], obj.get("id"))
        meta = from_union([Meta.from_dict, from_none], obj.get("meta"))
        name = from_union([Name.from_dict, from_none], obj.get("name"))
        nick_name = from_union([from_str, from_none], obj.get("nickName"))
        phone_numbers = from_union([lambda x: from_list(PhoneNumberElement.from_dict, x), from_none], obj.get("phoneNumbers"))
        photos = from_union([lambda x: from_list(Photo.from_dict, x), from_none], obj.get("photos"))
        profile_url = from_union([from_str, from_none], obj.get("profileUrl"))
        roles = from_union([lambda x: from_list(PhoneNumberElement.from_dict, x), from_none], obj.get("roles"))
        schemas = from_union([lambda x: from_list(from_str, x), from_none], obj.get("schemas"))
        timezone = from_union([from_str, from_none], obj.get("timezone"))
        title = from_union([from_str, from_none], obj.get("title"))
        urn_scim_schemas_extension_enterprise_10 = from_union([UrnScimSchemasExtensionEnterprise10.from_dict, from_none], obj.get("urn:scim:schemas:extension:enterprise:1.0"))
        user_name = from_union([from_str, from_none], obj.get("userName"))
        return Resource(active, addresses, display_name, emails, external_id, groups, id, meta, name, nick_name, phone_numbers, photos, profile_url, roles, schemas, timezone, title, urn_scim_schemas_extension_enterprise_10, user_name)

    def to_dict(self) -> dict:
        result: dict = {}
        result["active"] = from_union([from_bool, from_none], self.active)
        result["addresses"] = from_union([lambda x: from_list(lambda x: to_class(Address, x), x), from_none], self.addresses)
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["emails"] = from_union([lambda x: from_list(lambda x: to_class(PhoneNumberElement, x), x), from_none], self.emails)
        result["externalId"] = from_union([from_str, from_none], self.external_id)
        result["groups"] = from_union([lambda x: from_list(lambda x: to_class(Group, x), x), from_none], self.groups)
        result["id"] = from_union([from_str, from_none], self.id)
        result["meta"] = from_union([lambda x: to_class(Meta, x), from_none], self.meta)
        result["name"] = from_union([lambda x: to_class(Name, x), from_none], self.name)
        result["nickName"] = from_union([from_str, from_none], self.nick_name)
        result["phoneNumbers"] = from_union([lambda x: from_list(lambda x: to_class(PhoneNumberElement, x), x), from_none], self.phone_numbers)
        result["photos"] = from_union([lambda x: from_list(lambda x: to_class(Photo, x), x), from_none], self.photos)
        result["profileUrl"] = from_union([from_str, from_none], self.profile_url)
        result["roles"] = from_union([lambda x: from_list(lambda x: to_class(PhoneNumberElement, x), x), from_none], self.roles)
        result["schemas"] = from_union([lambda x: from_list(from_str, x), from_none], self.schemas)
        result["timezone"] = from_union([from_str, from_none], self.timezone)
        result["title"] = from_union([from_str, from_none], self.title)
        result["urn:scim:schemas:extension:enterprise:1.0"] = from_union([lambda x: to_class(UrnScimSchemasExtensionEnterprise10, x), from_none], self.urn_scim_schemas_extension_enterprise_10)
        result["userName"] = from_union([from_str, from_none], self.user_name)
        return result


class Users:
    active: Optional[bool]
    display_name: Optional[str]
    emails: Optional[List[UsersEmail]]
    external_id: Optional[str]
    groups: Optional[List[str]]
    id: Optional[str]
    items_per_page: Optional[int]
    meta: Optional[Meta]
    name: Optional[Name]
    nick_name: Optional[str]
    photos: Optional[List[Photo]]
    profile_url: Optional[str]
    resources: Optional[List[Resource]]
    schemas: Optional[List[str]]
    start_index: Optional[int]
    timezone: Optional[str]
    title: Optional[str]
    total_results: Optional[int]
    user_name: Optional[str]

    def __init__(self, active: Optional[bool], display_name: Optional[str], emails: Optional[List[UsersEmail]], external_id: Optional[str], groups: Optional[List[str]], id: Optional[str], items_per_page: Optional[int], meta: Optional[Meta], name: Optional[Name], nick_name: Optional[str], photos: Optional[List[Photo]], profile_url: Optional[str], resources: Optional[List[Resource]], schemas: Optional[List[str]], start_index: Optional[int], timezone: Optional[str], title: Optional[str], total_results: Optional[int], user_name: Optional[str]) -> None:
        self.active = active
        self.display_name = display_name
        self.emails = emails
        self.external_id = external_id
        self.groups = groups
        self.id = id
        self.items_per_page = items_per_page
        self.meta = meta
        self.name = name
        self.nick_name = nick_name
        self.photos = photos
        self.profile_url = profile_url
        self.resources = resources
        self.schemas = schemas
        self.start_index = start_index
        self.timezone = timezone
        self.title = title
        self.total_results = total_results
        self.user_name = user_name

    @staticmethod
    def from_dict(obj: Any) -> 'Users':
        assert isinstance(obj, dict)
        active = from_union([from_bool, from_none], obj.get("active"))
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        emails = from_union([lambda x: from_list(UsersEmail.from_dict, x), from_none], obj.get("emails"))
        external_id = from_union([from_str, from_none], obj.get("externalId"))
        groups = from_union([lambda x: from_list(from_str, x), from_none], obj.get("groups"))
        id = from_union([from_str, from_none], obj.get("id"))
        items_per_page = from_union([from_int, from_none], obj.get("itemsPerPage"))
        meta = from_union([Meta.from_dict, from_none], obj.get("meta"))
        name = from_union([Name.from_dict, from_none], obj.get("name"))
        nick_name = from_union([from_str, from_none], obj.get("nickName"))
        photos = from_union([lambda x: from_list(Photo.from_dict, x), from_none], obj.get("photos"))
        profile_url = from_union([from_str, from_none], obj.get("profileUrl"))
        resources = from_union([lambda x: from_list(Resource.from_dict, x), from_none], obj.get("Resources"))
        schemas = from_union([lambda x: from_list(from_str, x), from_none], obj.get("schemas"))
        start_index = from_union([from_int, from_none], obj.get("startIndex"))
        timezone = from_union([from_str, from_none], obj.get("timezone"))
        title = from_union([from_str, from_none], obj.get("title"))
        total_results = from_union([from_int, from_none], obj.get("totalResults"))
        user_name = from_union([from_str, from_none], obj.get("userName"))
        return Users(active, display_name, emails, external_id, groups, id, items_per_page, meta, name, nick_name, photos, profile_url, resources, schemas, start_index, timezone, title, total_results, user_name)

    def to_dict(self) -> dict:
        result: dict = {}
        result["active"] = from_union([from_bool, from_none], self.active)
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["emails"] = from_union([lambda x: from_list(lambda x: to_class(UsersEmail, x), x), from_none], self.emails)
        result["externalId"] = from_union([from_str, from_none], self.external_id)
        result["groups"] = from_union([lambda x: from_list(from_str, x), from_none], self.groups)
        result["id"] = from_union([from_str, from_none], self.id)
        result["itemsPerPage"] = from_union([from_int, from_none], self.items_per_page)
        result["meta"] = from_union([lambda x: to_class(Meta, x), from_none], self.meta)
        result["name"] = from_union([lambda x: to_class(Name, x), from_none], self.name)
        result["nickName"] = from_union([from_str, from_none], self.nick_name)
        result["photos"] = from_union([lambda x: from_list(lambda x: to_class(Photo, x), x), from_none], self.photos)
        result["profileUrl"] = from_union([from_str, from_none], self.profile_url)
        result["Resources"] = from_union([lambda x: from_list(lambda x: to_class(Resource, x), x), from_none], self.resources)
        result["schemas"] = from_union([lambda x: from_list(from_str, x), from_none], self.schemas)
        result["startIndex"] = from_union([from_int, from_none], self.start_index)
        result["timezone"] = from_union([from_str, from_none], self.timezone)
        result["title"] = from_union([from_str, from_none], self.title)
        result["totalResults"] = from_union([from_int, from_none], self.total_results)
        result["userName"] = from_union([from_str, from_none], self.user_name)
        return result


def users_from_dict(s: Any) -> Users:
    return Users.from_dict(s)


def users_to_dict(x: Users) -> Any:
    return to_class(Users, x)



# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = group_from_dict(json.loads(json_string))

from typing import Optional, Any, List, TypeVar, Type, cast, Callable


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


class Meta:
    created: Optional[str]
    location: Optional[str]

    def __init__(self, created: Optional[str], location: Optional[str]) -> None:
        self.created = created
        self.location = location

    @staticmethod
    def from_dict(obj: Any) -> 'Meta':
        assert isinstance(obj, dict)
        created = from_union([from_str, from_none], obj.get("created"))
        location = from_union([from_str, from_none], obj.get("location"))
        return Meta(created, location)

    def to_dict(self) -> dict:
        result: dict = {}
        result["created"] = from_union([from_str, from_none], self.created)
        result["location"] = from_union([from_str, from_none], self.location)
        return result


class Name:
    family_name: Optional[str]
    given_name: Optional[str]
    honorific_prefix: Optional[str]

    def __init__(self, family_name: Optional[str], given_name: Optional[str], honorific_prefix: Optional[str]) -> None:
        self.family_name = family_name
        self.given_name = given_name
        self.honorific_prefix = honorific_prefix

    @staticmethod
    def from_dict(obj: Any) -> 'Name':
        assert isinstance(obj, dict)
        family_name = from_union([from_str, from_none], obj.get("familyName"))
        given_name = from_union([from_str, from_none], obj.get("givenName"))
        honorific_prefix = from_union([from_str, from_none], obj.get("honorificPrefix"))
        return Name(family_name, given_name, honorific_prefix)

    def to_dict(self) -> dict:
        result: dict = {}
        result["familyName"] = from_union([from_str, from_none], self.family_name)
        result["givenName"] = from_union([from_str, from_none], self.given_name)
        result["honorificPrefix"] = from_union([from_str, from_none], self.honorific_prefix)
        return result


class Manager:
    manager_id: Optional[str]

    def __init__(self, manager_id: Optional[str]) -> None:
        self.manager_id = manager_id

    @staticmethod
    def from_dict(obj: Any) -> 'Manager':
        assert isinstance(obj, dict)
        manager_id = from_union([from_str, from_none], obj.get("managerId"))
        return Manager(manager_id)

    def to_dict(self) -> dict:
        result: dict = {}
        result["managerId"] = from_union([from_str, from_none], self.manager_id)
        return result


class UrnScimSchemasExtensionEnterprise10:
    cost_center: Optional[str]
    department: Optional[str]
    division: Optional[str]
    employee_number: Optional[str]
    manager: Optional[Manager]
    organization: Optional[str]

    def __init__(self, cost_center: Optional[str], department: Optional[str], division: Optional[str], employee_number: Optional[str], manager: Optional[Manager], organization: Optional[str]) -> None:
        self.cost_center = cost_center
        self.department = department
        self.division = division
        self.employee_number = employee_number
        self.manager = manager
        self.organization = organization

    @staticmethod
    def from_dict(obj: Any) -> 'UrnScimSchemasExtensionEnterprise10':
        assert isinstance(obj, dict)
        cost_center = from_union([from_str, from_none], obj.get("costCenter"))
        department = from_union([from_str, from_none], obj.get("department"))
        division = from_union([from_str, from_none], obj.get("division"))
        employee_number = from_union([from_str, from_none], obj.get("employeeNumber"))
        manager = from_union([Manager.from_dict, from_none], obj.get("manager"))
        organization = from_union([from_str, from_none], obj.get("organization"))
        return UrnScimSchemasExtensionEnterprise10(cost_center, department, division, employee_number, manager, organization)

    def to_dict(self) -> dict:
        result: dict = {}
        result["costCenter"] = from_union([from_str, from_none], self.cost_center)
        result["department"] = from_union([from_str, from_none], self.department)
        result["division"] = from_union([from_str, from_none], self.division)
        result["employeeNumber"] = from_union([from_str, from_none], self.employee_number)
        result["manager"] = from_union([lambda x: to_class(Manager, x), from_none], self.manager)
        result["organization"] = from_union([from_str, from_none], self.organization)
        return result


class Member:
    active: Optional[bool]
    display_name: Optional[str]
    external_id: Optional[str]
    id: Optional[str]
    locale: Optional[str]
    meta: Optional[Meta]
    name: Optional[Name]
    nick_name: Optional[str]
    password: Optional[str]
    preferred_language: Optional[str]
    profile_url: Optional[str]
    schemas: Optional[List[str]]
    timezone: Optional[str]
    title: Optional[str]
    urn_scim_schemas_extension_enterprise_10: Optional[UrnScimSchemasExtensionEnterprise10]
    user_name: Optional[str]
    user_type: Optional[str]

    def __init__(self, active: Optional[bool], display_name: Optional[str], external_id: Optional[str], id: Optional[str], locale: Optional[str], meta: Optional[Meta], name: Optional[Name], nick_name: Optional[str], password: Optional[str], preferred_language: Optional[str], profile_url: Optional[str], schemas: Optional[List[str]], timezone: Optional[str], title: Optional[str], urn_scim_schemas_extension_enterprise_10: Optional[UrnScimSchemasExtensionEnterprise10], user_name: Optional[str], user_type: Optional[str]) -> None:
        self.active = active
        self.display_name = display_name
        self.external_id = external_id
        self.id = id
        self.locale = locale
        self.meta = meta
        self.name = name
        self.nick_name = nick_name
        self.password = password
        self.preferred_language = preferred_language
        self.profile_url = profile_url
        self.schemas = schemas
        self.timezone = timezone
        self.title = title
        self.urn_scim_schemas_extension_enterprise_10 = urn_scim_schemas_extension_enterprise_10
        self.user_name = user_name
        self.user_type = user_type

    @staticmethod
    def from_dict(obj: Any) -> 'Member':
        assert isinstance(obj, dict)
        active = from_union([from_bool, from_none], obj.get("active"))
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        external_id = from_union([from_str, from_none], obj.get("externalId"))
        id = from_union([from_str, from_none], obj.get("id"))
        locale = from_union([from_str, from_none], obj.get("locale"))
        meta = from_union([Meta.from_dict, from_none], obj.get("meta"))
        name = from_union([Name.from_dict, from_none], obj.get("name"))
        nick_name = from_union([from_str, from_none], obj.get("nickName"))
        password = from_union([from_str, from_none], obj.get("password"))
        preferred_language = from_union([from_str, from_none], obj.get("preferredLanguage"))
        profile_url = from_union([from_str, from_none], obj.get("profileUrl"))
        schemas = from_union([lambda x: from_list(from_str, x), from_none], obj.get("schemas"))
        timezone = from_union([from_str, from_none], obj.get("timezone"))
        title = from_union([from_str, from_none], obj.get("title"))
        urn_scim_schemas_extension_enterprise_10 = from_union([UrnScimSchemasExtensionEnterprise10.from_dict, from_none], obj.get("urn:scim:schemas:extension:enterprise:1.0"))
        user_name = from_union([from_str, from_none], obj.get("userName"))
        user_type = from_union([from_str, from_none], obj.get("userType"))
        return Member(active, display_name, external_id, id, locale, meta, name, nick_name, password, preferred_language, profile_url, schemas, timezone, title, urn_scim_schemas_extension_enterprise_10, user_name, user_type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["active"] = from_union([from_bool, from_none], self.active)
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["externalId"] = from_union([from_str, from_none], self.external_id)
        result["id"] = from_union([from_str, from_none], self.id)
        result["locale"] = from_union([from_str, from_none], self.locale)
        result["meta"] = from_union([lambda x: to_class(Meta, x), from_none], self.meta)
        result["name"] = from_union([lambda x: to_class(Name, x), from_none], self.name)
        result["nickName"] = from_union([from_str, from_none], self.nick_name)
        result["password"] = from_union([from_str, from_none], self.password)
        result["preferredLanguage"] = from_union([from_str, from_none], self.preferred_language)
        result["profileUrl"] = from_union([from_str, from_none], self.profile_url)
        result["schemas"] = from_union([lambda x: from_list(from_str, x), from_none], self.schemas)
        result["timezone"] = from_union([from_str, from_none], self.timezone)
        result["title"] = from_union([from_str, from_none], self.title)
        result["urn:scim:schemas:extension:enterprise:1.0"] = from_union([lambda x: to_class(UrnScimSchemasExtensionEnterprise10, x), from_none], self.urn_scim_schemas_extension_enterprise_10)
        result["userName"] = from_union([from_str, from_none], self.user_name)
        result["userType"] = from_union([from_str, from_none], self.user_type)
        return result


class Group:
    display_name: Optional[str]
    id: Optional[str]
    members: Optional[List[Member]]
    meta: Optional[Meta]
    schemas: Optional[List[str]]

    def __init__(self, display_name: Optional[str], id: Optional[str], members: Optional[List[Member]], meta: Optional[Meta], schemas: Optional[List[str]]) -> None:
        self.display_name = display_name
        self.id = id
        self.members = members
        self.meta = meta
        self.schemas = schemas

    @staticmethod
    def from_dict(obj: Any) -> 'Group':
        assert isinstance(obj, dict)
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        id = from_union([from_str, from_none], obj.get("id"))
        members = from_union([lambda x: from_list(Member.from_dict, x), from_none], obj.get("members"))
        meta = from_union([Meta.from_dict, from_none], obj.get("meta"))
        schemas = from_union([lambda x: from_list(from_str, x), from_none], obj.get("schemas"))
        return Group(display_name, id, members, meta, schemas)

    def to_dict(self) -> dict:
        result: dict = {}
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["id"] = from_union([from_str, from_none], self.id)
        result["members"] = from_union([lambda x: from_list(lambda x: to_class(Member, x), x), from_none], self.members)
        result["meta"] = from_union([lambda x: to_class(Meta, x), from_none], self.meta)
        result["schemas"] = from_union([lambda x: from_list(from_str, x), from_none], self.schemas)
        return result


def group_from_dict(s: Any) -> Group:
    return Group.from_dict(s)


def group_to_dict(x: Group) -> Any:
    return to_class(Group, x)



from typing import Dict


class SCIMRequest():
    def __init__(
        self,
        *,
        token: str,
        http_method: str,
        url: str,
        query_params: Dict[str, str] = dict(),
        body_params: Dict[str, str] = dict(),
        json_body: Dict = dict(),
        headers: Dict[str, str] = dict(),
    ):
        self.token = token
        self.http_method = http_method
        self.url = url
        self.headers = headers
        self.query_params = query_params
        self.body_params = body_params
        self.json_body = json_body

    def to_dict(self) -> dict:
        result: dict = {"token": "redacted"}
        result["http_method"] = self.http_method
        result["url"] = self.url
        result["headers"] = self.headers
        result["query_params"] = self.query_params
        result["body_params"] = self.body_params
        result["json_body"] = self.json_body
        return result

    def __str__(self):
        return str(self.to_dict())



from typing import Dict


class SCIMResponse():
    def __init__(
        self,
        *,
        status: int,
        reason: str,
        headers: Dict[str, str],
        body: str,
    ):
        self.status = status
        self.reason = reason
        self.headers = headers
        self.body = body

    def is_success(self) -> bool:
        return self.status < 300

    def to_dict(self) -> dict:
        result: dict = {}
        result["status"] = self.status
        result["reason"] = self.reason
        result["headers"] = self.headers
        result["body"] = self.body
        return result

    def __str__(self):
        return str(self.to_dict())
